#include <cstdint>
#include <iostream>
#include <cassert>
#include <cstdio>
#include <vector>
#include <complex>
#include <random>

#include "cufft.h"

#define CUDA_CHECK(e)                                                            \
{                                                                                \
    cudaError_t err = (e);                                                       \
    if (err != cudaSuccess) {                                                    \
        fprintf(stderr, "CUDA error: %s, line %d, %s: %s\n", __FILE__, __LINE__, \
                #e, cudaGetErrorString(err));                                    \
        exit(EXIT_FAILURE);                                                      \
    }                                                                                \
}

#define CUFFT_CHECK(e)                                                         \
{                                                                              \
    cufftResult err = (e);                                                     \
    if (err != CUFFT_SUCCESS) {                                                \
        fprintf(stderr, "CUFFT error: %s, line %d, %s: %i\n", __FILE__,          \
                __LINE__, #e, err);                                              \
        exit(EXIT_FAILURE);                                                      \
    }                                                                          \
}

int main(int argc, char** argv) {

    size_t N = 128;
    assert(N % 2 == 0);

    /** Generate some real input signal **/

    std::vector<std::complex<float>> signal(N);
    for(size_t i = 0; i < N; i++) {
        // printf("signal[%zu] %f %f\n", i, signal[i].real(), signal[i].imag());
        long long k = 5*N/8;
        long long s = (N/4);
        signal[i] = std::complex<float>(std::exp(std::abs(k-(long long)i)/s), 0);
    }

    /** Compute DFT(signal)
     * Because signal is real, 
     * - input[0] is real
     * - input[N/2] is real
     * - input[i] = conj(input[N-i])
     * See https://en.wikipedia.org/wiki/Discrete_Fourier_transform#DFT_of_real_and_purely_imaginary_signals
     */

    std::vector<std::complex<float>> input(N);
    {
        cufftHandle plan;
        cufftComplex *in, *out;
        CUDA_CHECK(cudaMalloc(&in,  sizeof(cufftComplex) * N));
        CUDA_CHECK(cudaMalloc(&out, sizeof(cufftComplex) * N));
        CUDA_CHECK(cudaMemcpy(in, signal.data(), sizeof(cufftComplex) * N, cudaMemcpyDefault));
        CUFFT_CHECK(cufftPlan1d(&plan, N, CUFFT_C2C, 1));
        CUFFT_CHECK(cufftExecC2C(plan, in, out, CUFFT_FORWARD));
        CUFFT_CHECK(cufftDestroy(plan));
        CUDA_CHECK(cudaMemcpy(input.data(), out, sizeof(cufftComplex) * N, cudaMemcpyDefault));
        CUDA_CHECK(cudaFree(in));
        CUDA_CHECK(cudaFree(out));

        // Check symmetry
        float error = 0;
        float norm = 0;
        for(size_t i = 0; i < N; i++) {
            // printf("input[%zu] %f %f\n", i, input[i].real(), input[i].imag());
            size_t mirror = (N-i) % N;
            error += std::abs(input[i] - std::conj(input[mirror]));
            norm += std::abs(input[i]);
        }
        printf("Symmetry error: %e\n", error / norm);
    }

    /** Compute C2R transform using shorter complex input of length N/2+1
     * Note that the complex input of length N/2+1 needs to exactly satisfy the
     * properties from above, namely
     * - input_c2r[0] is real
     * - input_c2r[N/2] is real
     */
    for(int mistake = 0; mistake < 2; mistake++) {

        std::vector<std::complex<float>> input_c2r(N/2+1);
        std::vector<float> output_c2r(N);

        {
            for(size_t i = 0; i < N/2+1; i++) {
                input_c2r[i] = input[i];
            }

            // Add random garbage in the last entry - this is a mistake
            // even though the last entry's imaginary part is supposed to be zero
            // some algorithm read it
            if(mistake) {
                input_c2r.back().imag(5.0f);
                printf("Last value of input_c2r set to {%f,%f} instead of {%f, 0}\n", input_c2r.back().real(), input_c2r.back().imag(), input_c2r.back().real());
            }

            // We need a large batch to trigger kernels relying on the symmetry property (most don't)
            // We ignore whatever is computed in all batches except batch 0
            size_t batch = 8192;
            cufftHandle plan;
            cufftComplex *in;
            float *out;
            CUDA_CHECK(cudaMalloc(&in,  batch * sizeof(cufftComplex) * (N/2+1)));
            CUDA_CHECK(cudaMalloc(&out, batch * sizeof(float) *         N));
            CUDA_CHECK(cudaMemset(in, batch * sizeof(cufftComplex) * (N/2+1), 3));
            CUDA_CHECK(cudaMemcpy(in, input_c2r.data(), sizeof(cufftComplex) * (N/2+1), cudaMemcpyDefault));
            CUFFT_CHECK(cufftPlan1d(&plan, N, CUFFT_C2R, batch));
            CUFFT_CHECK(cufftExecC2R(plan, in, out));
            CUFFT_CHECK(cufftDestroy(plan));
            CUDA_CHECK(cudaMemcpy(output_c2r.data(), out, sizeof(float) * N, cudaMemcpyDefault));
            CUDA_CHECK(cudaFree(in));
            CUDA_CHECK(cudaFree(out));

            // Add normalization factor
            for(auto &v: output_c2r) {
                v /= N;
            }

            float error = 0;
            float norm = 0;
            for(size_t i = 0; i < N; i++) {
                error += std::abs(output_c2r[i] - signal[i].real());
                norm += std::abs(signal[i].real());
            }
            printf("Error C2R output vs C2C output (with mistake ? %d): %e\n", mistake, error / norm);
        }
    }
}